from typing import Annotated

import numpy as np
from gpxpy.gpx import GPXTrackPoint
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic_numpy import np_array_pydantic_annotated_typing

from geo_track_analyzer.utils.internal import GPXTrackPointAfterValidator


class Model(BaseModel):
    model_config = ConfigDict(extra="forbid")


class Position2D(Model):
    """Position in a 2D latitude / longitude space"""

    latitude: float
    """Latitude of the point"""

    longitude: float
    """Longitude of the point"""


class Position3D(Position2D):
    """Position in a 3D latitude / longitude / elevation space"""

    elevation: None | float
    """Elevation of the point"""


class ElevationMetrics(Model):
    """Collection of elevation related metrics"""

    uphill: float
    """Upill elevation in m"""

    downhill: float
    """Downhill elevation in m"""

    slopes: list[float]
    """Slopes between points in a uphill/downhill section"""


class SegmentOverview(Model):
    """Collection of metrics for a Segment"""

    moving_time_seconds: float
    """Moving time of a segment"""

    total_time_seconds: float
    """Total time of the segment"""

    moving_distance: float
    """Moving distance (point-to-point distance with velocity below a threshold) in m"""

    total_distance: float
    """Total distance of the segment in m"""

    max_velocity: None | float
    """Maximum velocity in the segment in m/s (only considering velocities below the XX
    percentile)"""

    avg_velocity: None | float
    """Average velocity in the segment in m/s (only considering velocities below the XX
    percentile)"""

    max_elevation: None | float
    """Maximum elevation in the segment in m"""

    min_elevation: None | float
    """ Minimum elevation in the segment in m"""

    uphill_elevation: None | float
    """Elevation traveled uphill in m"""

    downhill_elevation: None | float
    """Elevation traveled downhill in m"""

    # Attributes that will be calculated from primary attributes
    moving_distance_km: float = Field(default=-1)
    """moving_distance converted the km"""

    total_distance_km: float = Field(default=-1)
    """total_distance converted the km"""

    max_velocity_kmh: None | float = Field(default=None)
    """max_velocity converted the km/h"""

    avg_velocity_kmh: None | float = Field(default=None)
    """avg_speed converted the km/h"""

    @model_validator(mode="after")
    def set_km_attr(self) -> "SegmentOverview":
        self.moving_distance_km = self.moving_distance / 1000
        self.total_distance_km = self.total_distance / 1000
        self.max_velocity_kmh = (
            None if self.max_velocity is None else 3.6 * self.max_velocity
        )
        self.avg_velocity_kmh = (
            None if self.avg_velocity is None else 3.6 * self.avg_velocity
        )
        return self


class SegmentOverlap(Model):
    """Represent the overlap between two segments"""

    overlap: float
    """Overlap ratio of the segments"""

    inverse: bool
    """Match direction of the segment relative to the base"""

    plate: np_array_pydantic_annotated_typing(data_type=np.float32, dimensions=2)
    """2D representation of the segment overlap"""

    start_point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """First point matching the base segment """

    start_idx: int
    """Index of the first point in match segment"""

    end_point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """Last point matching the base segment """

    end_idx: int
    """Index of the last point in match segment"""

    def __repr__(self) -> str:
        ret_str = f"Overlap {self.overlap*100:.2f}%, Inverse: {self.inverse},"
        ret_str += f" Plate: {self.plate.shape}, Points: "
        point_strs = []
        for point, idx in zip(
            [self.start_point, self.end_point], [self.start_idx, self.end_idx]
        ):
            point_strs.append(f"({point.latitude},{point.longitude}) at id {idx}")
        ret_str += " to ".join(point_strs)

        return ret_str


class PointDistance(Model):
    """Represents a distance calculati  on to a point in a GPX track."""

    point: Annotated[GPXTrackPoint, GPXTrackPointAfterValidator]
    """The nearest point on the track."""

    distance: float
    """The distance to the nearest point."""

    point_idx_abs: int
    """The absolute index of the nearest point in the track."""

    segment_idx: int
    """The index of the segment containing the nearest point."""

    segment_point_idx: int
    """The index of the nearest point within the containing segment."""

    def __repr__(self) -> str:
        return (
            f"PointDistance(point=({self.point.latitude},{self.point.longitude})"
            f"distance={self.distance}, "
            f"point_idx_abs={self.point_idx_abs}, "
            f"segment_idx={self.segment_idx}, "
            f"segment_point_idx={self.segment_point_idx})"
        )



import logging

import pandas as pd
import plotly.express as px
from plotly.graph_objs import Figure

logger = logging.getLogger(__name__)


def plot_track_3d(data: pd.DataFrame, strict_data_selection: bool = False) -> Figure:
    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot = data[mask]

    return px.line_3d(data_for_plot, x="latitude", y="longitude", z="elevation")



class APIResponseError(Exception):
    pass


class APIHealthCheckFailedError(Exception):
    pass


class APIDataNotAvailableError(Exception):
    pass


class TrackInitializationError(Exception):
    pass


class TrackTransformationError(Exception):
    pass


class InvalidBoundsError(Exception):
    pass


class GPXPointExtensionError(Exception):
    pass


class VisualizationSetupError(Exception):
    pass


class TrackAnalysisError(Exception):
    pass



import logging

import pandas as pd
import plotly.graph_objects as go
from plotly.graph_objs import Figure
from plotly.subplots import make_subplots

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.visualize.utils import get_slope_colors

logger = logging.getLogger(__name__)


def plot_track_2d(
    data: pd.DataFrame,
    *,
    include_velocity: bool = False,
    include_heartrate: bool = False,
    include_cadence: bool = False,
    include_power: bool = False,
    strict_data_selection: bool = False,
    height: None | int = 600,
    width: None | int = 1800,
    pois: None | list[tuple[float, float]] = None,
    color_elevation: None | str = None,
    color_additional_trace: None | str = None,
    color_poi: None | str = None,
    slider: bool = False,
    **kwargs,
) -> Figure:
    """Elevation profile of the track. May be enhanced with additional information like
    Velocity, Heartrate, Cadence, and Power.

    :param data: DataFrame containing track data
    :param include_velocity: Plot velocity as second y-axis, defaults to False
    :param include_heartrate: Plot heart rate as second y-axis, defaults to False
    :param include_cadence: Plot cadence as second y-axis, defaults to False
    :param include_power: Plot power as second y-axis, defaults to False
    :param strict_data_selection: If True only included that passing the minimum speed
        requirements of the Track, defaults to False
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1800
    :param pois: Optional lat/long coordingates to add to the plot as points of
        interest, defaults to None
    :param color_elevation: Color of the elevation as str interpretable by plotly,
        defaults to None
    :param color_additional_trace: Color of velocity/heartrate/cadence/power as str
        interpretable by plotly, defaults to None
    :param color_poi: Color of the pois as str interpretable by plotly, defaults to None
    :param slider: Should a slide be included in the plot to zoom into the x-axis,
        defaults to False
    :raises VisualizationSetupError: If more than one of include_velocity,
        include_heartrate, include_cadence, or include_power was set the True
    :raises VisualizationSetupError: If elevation data is missing in the data
    :raises VisualizationSetupError: If the data requried for the additional data is
       missing

    :return: Plotly Figure object.
    """
    if (
        sum(
            [
                int(include_velocity),
                int(include_heartrate),
                int(include_cadence),
                int(include_power),
            ]
        )
        > 1
    ):
        raise VisualizationSetupError(
            "Only one of include_velocity, include_heartrate, include_cadence, "
            "and include_power can be set to True"
        )

    mask = data.moving
    if strict_data_selection:
        mask = mask & data.in_speed_percentile

    data_for_plot = data[mask]

    if data_for_plot.elevation.isna().all():
        raise VisualizationSetupError("Can not plot profile w/o elevation information")

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    fig.add_trace(
        go.Scatter(
            x=data_for_plot.cum_distance_moving,
            y=data_for_plot.elevation,
            mode="lines",
            name="Elevation [m]",
            fill="tozeroy",
        ),
        secondary_y=False,
    )
    fig.update_yaxes(
        title_text="Elevation [m]",
        secondary_y=False,
        range=[
            data_for_plot.elevation.min() * 0.97,
            data_for_plot.elevation.max() * 1.05,
        ],
    )
    fig.update_xaxes(title_text="Distance [m]")

    y_data = None
    y_range = None
    title = None
    mode = "lines"
    fill: None | str = "tozeroy"
    if include_velocity:
        y_data = data_for_plot.apply(lambda c: c.speed * 3.6, axis=1)
        title = "Velocity [km/h]"
        y_range = [0, y_data.max() * 2.1]
    if include_heartrate:
        if pd.isna(data_for_plot.heartrate).all():
            raise VisualizationSetupError(
                "Requested to plot heart rate but no heart rate information available "
                "in data"
            )
        y_data = data_for_plot.heartrate.fillna(0).astype(int)
        title = "Heart Rate [bpm]"
        y_range = [0, y_data.max() * 1.2]
    if include_cadence:
        if pd.isna(data_for_plot.cadence).all():
            raise VisualizationSetupError(
                "Requested to plot cadence but no cadence information available in data"
            )
        y_data = data_for_plot.cadence.fillna(0).astype(int)
        title = "Cadence [rpm]"
        mode = "markers"
        fill = None
        y_range = [0, y_data.max() * 1.2]
    if include_power:
        if pd.isna(data_for_plot.power).all():
            raise VisualizationSetupError(
                "Requested to plot power but no power information available in data"
            )
        y_data = data_for_plot.power.fillna(0).astype(int)
        title = "Power [W]"
        y_range = [0, y_data.max() * 1.2]

    if y_data is not None:
        fig.add_trace(
            go.Scatter(
                x=data_for_plot.cum_distance_moving,
                y=y_data,
                mode=mode,
                name=title,
                fill=fill,
            ),
            secondary_y=True,
        )
        fig.update_yaxes(
            title_text=title,
            secondary_y=True,
            range=y_range,
        )

    if pois is not None:
        for i_poi, poi in enumerate(pois):
            lat, lng = poi
            poi_data = data_for_plot[
                (data_for_plot.latitude == lat) & (data_for_plot.longitude == lng)
            ]
            if poi_data.empty:
                logger.warning("Could not find POI in data. Skipping")
                continue
            poi_x = poi_data.iloc[0].cum_distance_moving
            poi_y = poi_data.iloc[0].elevation

            fig.add_scatter(
                name=f"PIO {i_poi} @ {lat} / {lng}",
                x=[poi_x],
                y=[poi_y],
                mode="markers",
                marker=dict(
                    size=20,
                    color="MediumPurple" if color_poi is None else color_poi,
                    symbol="triangle-up",
                    standoff=10,
                    angle=180,
                ),
            )

    fig.update_layout(
        showlegend=False, autosize=False, margin={"r": 0, "t": 0, "l": 0, "b": 0}
    )
    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    fig.update_xaxes(
        range=[
            data_for_plot.iloc[0].cum_distance_moving,
            data_for_plot.iloc[-1].cum_distance_moving,
        ]
    )

    if slider:
        fig.update_layout(
            xaxis=dict(
                rangeslider=dict(visible=True),
            )
        )

    if color_elevation is not None:
        fig.data[0].marker.color = color_elevation  # type: ignore
    if color_additional_trace is not None and any(
        [include_velocity, include_heartrate, include_cadence, include_power]
    ):
        fig.data[1].marker.color = color_additional_trace  # type: ignore

    return fig


def plot_track_with_slope(
    data: pd.DataFrame,
    *,
    slope_gradient_color: tuple[str, str, str] = ("#0000FF", "#00FF00", "#FF0000"),
    min_slope: int = -18,
    max_slope: int = 18,
    height: None | int = 600,
    width: None | int = 1800,
    slider: bool = False,
    **kwargs,
) -> Figure:
    """Elevation profile with slopes between points.

    :param data: DataFrame containing track data
    :param slope_gradient_color: Colors for the min, neutral, max slope values,
        defaults to ("#0000FF", "#00FF00", "#FF0000")
    :param min_slope: Minimum slope for the gradient also acts as floor for the
        displayed slope, defaults to -18
    :param max_slope: Maximum  slope for the gradient also acts as ceiling for the
        displayed slope, defaults to 18
    :param height: Height of the plot, defaults to 600
    :param width: Width of the plot, defaults to 1800
    :param slider: Should a slide be included in the plot to zoom into the x-axis,
        defaults to False
    :raises VisualizationSetupError: If elevation data is missing in the data

    :return: Plotly Figure object
    """
    slope_color_map = get_slope_colors(
        *slope_gradient_color, max_slope=max_slope, min_slope=min_slope
    )

    data = data[data.moving].copy()

    if data.elevation.isna().all():
        raise VisualizationSetupError("Can not plot profile w/o elevation information")

    elevations = data.elevation.to_list()
    diff_elevation = [0]
    for i, elevation in enumerate(elevations[1:]):
        diff_elevation.append(elevation - elevations[i])

    data["elevation_diff"] = diff_elevation

    def calc_slope(row: pd.Series) -> int:
        try:
            slope = round((row.elevation_diff / row.distance) * 100)
        except ZeroDivisionError:
            slope = 0

        if slope > max_slope:
            slope = max_slope
        elif slope < min_slope:
            slope = min_slope

        return slope

    data["slope"] = data.apply(lambda row: calc_slope(row), axis=1).astype(int)

    fig = go.Figure()

    fig.add_trace(
        go.Scatter(
            x=data.cum_distance_moving,
            y=data.elevation,
            mode="lines",
            name="Elevation [m]",
            fill="tozeroy",
        )
    )

    if slider:
        fig.update_layout(
            xaxis=dict(
                rangeslider=dict(visible=True),
            )
        )

    for i in range(len(data)):
        this_data = data.iloc[i : i + 2]
        if len(this_data) == 1:
            continue

        slope_val = this_data.iloc[1].slope

        color = slope_color_map[slope_val]
        max_distance: float = max(this_data.cum_distance_moving)
        fig.add_trace(
            go.Scatter(
                x=this_data.cum_distance_moving,
                y=this_data.elevation,
                mode="lines",
                name=f"Distance {max_distance/1000:.1f} km",
                fill="tozeroy",
                marker_color=color,
                hovertemplate=f"Slope: {slope_val} %",
            )
        )

    fig.update_layout(
        showlegend=False, autosize=False, margin={"r": 0, "t": 0, "l": 0, "b": 0}
    )

    if height is not None:
        fig.update_layout(height=height)
    if width is not None:
        fig.update_layout(width=width)

    min_elevation: float = min(data.elevation)
    max_elevation: float = max(data.elevation)
    fig.update_yaxes(
        showspikes=True,
        spikemode="across",
        range=[min_elevation * 0.95, max_elevation * 1.05],
        title_text="Elevation [m]",
    )
    fig.update_xaxes(title_text="Distance [m]")

    return fig



from typing import Literal

import pandas as pd

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.processing import (
    get_processed_segment_data,
    get_processed_track_data,
)
from geo_track_analyzer.track import Track, logger


def extract_track_data_for_plot(
    track: Track,
    kind: str,
    require_elevation: list[str],
    intervals: None | int = None,
    connect_segments: Literal["full", "forward"] = "forward",
) -> pd.DataFrame:
    """Extract the data from a Track as DataFrame for plotting.

    :param track: Track object
    :param kind: Kind of plot
    :param require_elevation: List of kinds of plots that require elevation data to be
        present in the Track
    :param intervals: Optionally reduce the pp-distance in the track, defaults to None

    :return: DataFrame
    """
    if kind in require_elevation and not track.track.has_elevations():
        raise VisualizationSetupError(f"Track has so elevation so {kind} is not valid")
    _track = track.track

    if intervals is not None:
        if track.get_avg_pp_distance() >= intervals:
            logger.debug("Average pp distance larget than interval. Skipping reduction")
        else:
            _track = _track.clone()
            _track.reduce_points(intervals)

    _, _, _, _, data = get_processed_track_data(
        _track, connect_segments=connect_segments
    )

    return data


def extract_segment_data_for_plot(
    track: Track,
    segment: int,
    kind: str,
    require_elevation: list[str],
    intervals: None | int = None,
) -> pd.DataFrame:
    """Extract the data for a segment from a Track as DataFrame for plotting.

    :param track: Track object
    :param segment: Index of the segment to be extracted
    :param kind: Kind of plot
    :param require_elevation: List of kinds of plots that require elevation data to be
        present in the Track
    :param intervals: Optionally reduce the pp-distance in the track, defaults to None

    :return: DataFrame
    """
    if kind in require_elevation and not track.track.segments[segment].has_elevations():
        raise VisualizationSetupError(
            f"Segment has so elevation so {kind} is not valid"
        )
    if kind == "map-segments":
        raise VisualizationSetupError("map-segments can only be done for full tracks")

    segement = track.track.segments[segment]

    if intervals is not None:
        if track.get_avg_pp_distance_in_segment(segment) >= intervals:
            logger.debug("Average pp distance larget than interval. Skipping reduction")
        else:
            segement = track.track.segments[segment].clone()
            segement.reduce_points(intervals)

    _, _, _, _, data = get_processed_segment_data(segement)
    return data



from .interactive import plot_track_3d
from .map import (
    plot_segments_on_map,
    plot_track_enriched_on_map,
    plot_track_line_on_map,
    plot_tracks_on_map,
)
from .profiles import plot_track_2d, plot_track_with_slope

__all__ = [
    "plot_track_3d",
    "plot_track_line_on_map",
    "plot_track_enriched_on_map",
    "plot_track_2d",
    "plot_track_with_slope",
    "plot_segments_on_map",
    "plot_tracks_on_map",
]



import logging
from datetime import timedelta
from typing import Literal

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.graph_objs import Figure

from geo_track_analyzer.exceptions import VisualizationSetupError
from geo_track_analyzer.utils.base import (
    center_geolocation,
    format_timedelta,
)
from geo_track_analyzer.visualize.constants import (
    COLOR_GRADIENTS,
    DEFAULT_COLOR_GRADIENT,
    ENRICH_UNITS,
)
from geo_track_analyzer.visualize.utils import get_color_gradient

logger = logging.getLogger(__name__)


def plot_track_line_on_map(
    data: pd.DataFrame,
    *,
    zoom: int = 13,
    height: None | int = None,
    width: None | int = None,
    line_width: float = 2.5,
    map_style: str = "open-street-map",
    **kwargs,
) -> Figure:
    """
    Plot the track line on a map.

    :param data: DataFrame containing latitude and longitude data.
    :param zoom: Zoom level for the map, defaults to 13
    :param height: Height of the plot, defaults to None
    :param width: Width of the plot, defaults to None
    :param line_width: Width of the line on the map, defaults to 2.5
    :param map_style: Valid map_style for plotly mapbox_style, defaults to
        open-street-map
    :param kwargs: Additional keyword arguments.

    :return: Plotly Figure object.
    """
    mask = data.moving

    center_lat, center_lon = center_geolocation(
        [(r["latitude"], r["longitude"]) for r in data[mask].to_dict("records")]
    )
    fig = px.line_mapbox(
        data[mask],
        lat="latitude",
        lon="longitude",
        zoom=zoom,
        center={"lon": center_lon, "lat": center_lat},
        height=height,
        width=width,
    )
    fig.update_traces(
        line=dict(width=line_width),
    )
    fig.update_layout(
        mapbox_style=map_style,
        margin={"r": 57, "t": 0, "l": 49, "b": 0},
    )

    return fig


def plot_track_enriched_on_map(
    data: pd.DataFrame,
    *,
    enrich_with_column: Literal[
        "elevation", "speed", "heartrate", "cadence", "power"
    ] = "elevation",
    zoom: int = 13,
    height: None | int = None,
    width: None | int = None,
    overwrite_color_gradient: None | tuple[str, str] = None,
    overwrite_unit_text: None | str = None,
    cbar_ticks: int = 5,
    map_style: str = "open-street-map",
    **kwargs,
) -> Figure:
    """
    Plot the track line enriched with additional information (as z-axis) on a map.

    :param data: DataFrame containing track data.
    :param enrich_with_column: Column to enrich the track with, defaults to "elevation"
    :param zoom: Zoom level for the map, defaults to 13
    :param height: Height of the plot, defaults to None
    :param width: Width of the plot, defaults to None
    :param overwrite_color_gradient: Custom color gradient for the plot. Check
        track_analyzer.visualize.constants for defaults, defaults to None
    :param overwrite_unit_text: Custom unit text for the enrichment. Check
        track_analyzer.visualize.constants for defaults, defaults to None
    :param cbar_ticks: Number of color bar ticks, defaults to 5
    :param map_style: Valid map_style for plotly mapbox_style, defaults to
        open-street-map
    :param kwargs: Additional keyword arguments.
    :raises VisualizationSetupError: If no data is in passed enrich_with_column column

    :return: Plotly Figure object.
    """
    mask = data.moving

    plot_data = data[mask]

    center_lat, center_lon = center_geolocation(
        [(r["latitude"], r["longitude"]) for r in data[mask].to_dict("records")]
    )

    # ~~~~~~~~~~~~ Enrichment data ~~~~~~~~~~~~~~~~
    enrich_unit = (
        ENRICH_UNITS[enrich_with_column]
        if overwrite_unit_text is None
        else overwrite_unit_text
    )
    enrich_type = (
        int if enrich_with_column in ["heartrate", "cadence", "power"] else float
    )

    # ~~~~~~~~~~~ Generator colors for passed column ~~~~~~~~~~~~~~~~~
    color_column_values = plot_data[enrich_with_column]

    if color_column_values.isna().all():
        raise VisualizationSetupError(
            f"Plotting not possible. No values for {enrich_with_column} in passed data."
        )

    if color_column_values.isna().any():
        logger.warning(
            "%s nan rows in %s. Dropping points",
            color_column_values.isna().sum(),
            enrich_with_column,
        )
        plot_data = plot_data[~color_column_values.isna()]
        color_column_values = color_column_values[~color_column_values.isna()]

    if enrich_with_column == "speed":
        color_column_values = color_column_values * 3.6
    diff_abs = color_column_values.max() - color_column_values.min()
    assert diff_abs > 0

    if overwrite_color_gradient:
        color_min, color_max = overwrite_color_gradient
    else:
        if enrich_with_column in COLOR_GRADIENTS.keys():
            color_min, color_max = COLOR_GRADIENTS[enrich_with_column]
        else:
            color_min, color_max = DEFAULT_COLOR_GRADIENT
    color_map = pd.Series(
        data=get_color_gradient(color_min, color_max, round(diff_abs) + 1),
        index=range(
            round(color_column_values.min()), round(color_column_values.max()) + 1
        ),
    )

    def color_mapper(value: float) -> str:
        if value < color_map.index.start:
            return color_map.iloc[0]
        elif value > color_map.index.stop:
            return color_map.iloc[-1]
        else:
            return color_map.loc[int(value)]

    colors = color_column_values.apply(color_mapper).to_list()
    marker = go.scattermapbox.Marker(color=colors)

    # ~~~~~~~~~~~~~~~ Colorbar for the passed column ~~~~~~~~~~~~~~~~~~~~
    splits = 1 / (cbar_ticks - 1)
    factor = 0.0
    tick_vals = []
    tick_cols = []
    while factor <= 1:
        idx = int(diff_abs * factor)
        tick_vals.append(color_map.index[idx])
        tick_cols.append(color_map.iloc[idx])
        factor += splits

    colorbar_trace = go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        marker=dict(
            colorscale=color_map.to_list(),
            showscale=True,
            cmin=color_column_values.min(),
            cmax=color_column_values.max(),
            colorbar=dict(
                title=enrich_with_column.capitalize(),
                thickness=10,
                tickvals=tick_vals,
                ticktext=tick_vals,
                outlinewidth=0,
            ),
        ),
        hoverinfo="none",
    )

    # ~~~~~~~~~~~~~~~ Build figure ~~~~~~~~~~~~~~~~~~~
    fig = go.Figure(
        go.Scattermapbox(
            lat=plot_data["latitude"],
            lon=plot_data["longitude"],
            mode="markers",
            marker=marker,
            hovertemplate=f"{enrich_with_column.capitalize()}: "
            + "<b>%{text}</b> "
            + f"{enrich_unit} <br>"
            + "<b>Lat</b>: %{lat:4.6f}°<br>"
            + "<b>Lon</b>: %{lon:4.6f}°<br>",
            text=[enrich_type(v) for v in color_column_values.to_list()],
            name="",
        )
    )

    fig.add_trace(colorbar_trace)

    fig.update_layout(mapbox_style=map_style)
    fig.update_layout(
        margin={"r": 57, "t": 5, "l": 49, "b": 5},
        mapbox={
            "style": map_style,
            "zoom": zoom,
            "center": {"lon": center_lon, "lat": center_lat},
        },
        height=height,
        hovermode="closest",
        width=width,
        showlegend=False,
    )

    return fig


def plot_segments_on_map(
    data: pd.DataFrame,
    *,
    zoom: int = 13,
    height: None | int = None,
    width: None | int = None,
    line_width: float = 2.5,
    average_only: bool = True,
    map_style: str = "open-street-map",
    **kwargs,
) -> Figure:
    """
    Plot track line on map. Segments are displayed individually.

    :param data: DataFrame containing track segment data.
    :param zoom: Zoom level for the map, defaults to 13
    :param height: Height of the plot, defaults to None
    :param width: Width of the plot, defaults to None
    :param line_width: Width of the line on the map, defaults to 2.5
    :param average_only: Flag to display averages only, defaults to True
    :param map_style: Valid map_style for plotly mapbox_style, defaults to
        open-street-map
    :param kwargs: Additional keyword arguments.
    :raises VisualizationSetupError: If no segment information is contained in the data
    :raises VisualizationSetupError: If there are not two or more segments in the data

    :return: Plotly Figure object.
    """
    mask = data.moving

    plot_data = data[mask]

    if "segment" not in plot_data.columns:
        raise VisualizationSetupError(
            "Data has no **segment** in columns. Required for plot"
        )
    if len(plot_data.segment.unique()) < 2:
        raise VisualizationSetupError("Data does not have mulitple segments")

    center_lat, center_lon = center_geolocation(
        [(r["latitude"], r["longitude"]) for r in data[mask].to_dict("records")]
    )

    fig = go.Figure()
    for i_segment, frame in plot_data.groupby(by="segment"):
        mean_heartrate = frame.heartrate.agg("mean")
        min_heartrate = frame.heartrate.agg("min")
        max_heartrate = frame.heartrate.agg("max")

        mean_speed = frame.speed.agg("mean") * 3.6
        min_speed = frame.speed.agg("min") * 3.6
        max_speed = frame.speed.agg("max") * 3.6

        mean_power = frame.power.agg("mean")
        min_power = frame.power.agg("min")
        max_power = frame.power.agg("max")

        distance = frame.distance.sum() / 1000
        if frame.time.isna().all():
            total_time = None
        else:
            _total_time = frame.time.sum()  # in seconds
            total_time = timedelta(seconds=int(_total_time.astype(int)))
        min_elevation = frame.elevation.min()
        max_elevation = frame.elevation.max()

        text: str = (
            f"<b>Segment {i_segment}</b><br>"
            + f"<b>Distance</b>: {distance:.2f} km<br>"
        )
        if total_time is not None:
            text += f"<b>Time</b>: {format_timedelta(total_time)}<br>"

        text += (
            f"<b>Elevation</b>: &#8600; {min_elevation} m "
            + f"&#8599; {max_elevation} m<br>"
        )
        if not np.isnan(mean_speed):
            text += f"<b>Speed</b>: &#248; {mean_speed:.1f} "
            if not average_only:
                text += f" &#8600;{min_speed:.1f} &#8599;{max_speed:.1f} km/h <br>"
            text += " km/h <br>"
        if not np.isnan(mean_heartrate):
            text += f"<b>Heartrate</b>: &#248; {int(mean_heartrate)} "
            if not average_only:
                text += f"&#8600;{int(min_heartrate)} &#8599;{int(max_heartrate)}<br>"
            text += " bpm<br>"
        if not np.isnan(mean_power):
            text += f"<b>Power</b>: &#248; {mean_power:.1f} "
            if not average_only:
                text += f"&#8600;{min_power:.1f} &#8599;{max_power:.1f}<br>"
            text += " W<br>"

        fig.add_trace(
            go.Scattermapbox(
                lat=frame["latitude"],
                lon=frame["longitude"],
                mode="lines",
                hovertemplate="%{text} ",
                text=len(frame) * [text],
                line=dict(width=line_width),
                name="",
            )
        )

    fig.update_layout(
        margin={"r": 57, "t": 5, "l": 49, "b": 5},
        mapbox={
            "style": map_style,
            "zoom": zoom,
            "center": {"lon": center_lon, "lat": center_lat},
        },
        height=height,
        hovermode="closest",
        width=width,
        showlegend=False,
    )

    return fig


def plot_tracks_on_map(
    datas: list[pd.DataFrame],
    names: list[str],
    *,
    zoom: int = 13,
    height: None | int = None,
    width: None | int = None,
    line_width: float = 2.5,
    colors: None | list[str] = None,
    map_style: str = "open-street-map",
    **kwargs,
) -> Figure:
    fig = go.Figure()

    all_points = []
    for i, (data, name) in enumerate(zip(datas, names)):
        mask = data.moving
        plot_data = data[mask]

        mod_line: dict[str, float | str | int] = {"width": line_width}
        if colors is not None:
            mod_line["color"] = colors[i]

        fig.add_trace(
            go.Scattermapbox(
                lat=plot_data["latitude"],
                lon=plot_data["longitude"],
                mode="lines",
                # hovertemplate="%{text} ",
                # text=len(frame) * [text],
                line=mod_line,
                name=name,
            )
        )
        all_points.extend(
            [(r["latitude"], r["longitude"]) for r in data[mask].to_dict("records")]
        )

    center_lat, center_lon = center_geolocation(all_points)

    fig.update_layout(
        margin={"r": 57, "t": 5, "l": 49, "b": 5},
        mapbox={
            "style": map_style,
            "zoom": zoom,
            "center": {"lon": center_lon, "lat": center_lat},
        },
        height=height,
        hovermode="closest",
        width=width,
        showlegend=False,
    )

    return fig



"""
Enhance gpx tracks with external data. E.g. elevation data
"""
import json
import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict, Mapping, Type, final

import requests
from gpxpy.gpx import GPXTrack
from requests.structures import CaseInsensitiveDict

from geo_track_analyzer.exceptions import (
    APIDataNotAvailableError,
    APIHealthCheckFailedError,
    APIResponseError,
)

logger = logging.getLogger(__name__)


class EnhancerType(str, Enum):
    OPENTOPOELEVATION = "OpenTopoElevation"
    OPENELEVATION = "OpenElevation"


class Enhancer(ABC):
    """Base class for GPX Track enhancement"""

    @abstractmethod
    def __init__(self, url: str) -> None:
        pass

    @abstractmethod
    def enhance_track(self, track: GPXTrack, inplace: bool = False) -> GPXTrack:
        pass


class ElevationEnhancer(Enhancer):
    """Base class for enhancing GPX Tracks with externally provided elevation data"""

    def enhance_track(self, track: GPXTrack, inplace: bool = False) -> GPXTrack:
        """
        Main method to enhance a passed GPX track with elevation information

        :param track: Track to be enhanced.

        :returns: The enhanced track
        """
        if inplace:
            track_ = track
        else:
            track_ = track.clone()

        for segment in track_.segments:
            request_coordinates = []
            for point in segment.points:
                request_coordinates.append((point.latitude, point.longitude))

            elevations = self.get_elevation_data(request_coordinates)
            for point, elevation in zip(segment.points, elevations):
                point.elevation = elevation

        return track_

    @abstractmethod
    def get_elevation_data(
        self, input_coordinates: list[tuple[float, float]]
    ) -> list[float]:
        pass


@final
class OpenTopoElevationEnhancer(ElevationEnhancer):
    """Use the/a OpenTopoData API (https://opentopodata.org) to enhance a GPX track
    with elevation information."""

    def __init__(
        self,
        url: str = "https://api.opentopodata.org/",
        dataset: str = "eudem25m",
        interpolation: str = "cubic",
        skip_checks: bool = False,
    ) -> None:
        """
        Setup the enhancer via a opentopodata rest api.

        :param url: REST api entrypoint url, defaults to "https://api.opentopodata.org/"
        :param dataset: Dataset of elevation data , defaults to "eudem25m"
        :param interpolation: Interpolation method, defaults to "cubic"
        :param skip_checks: If true, health checks will be skipped on initialization,
            defaults to False
        :raises APIHealthCheckFailedError: If connection can not established
        :raises APIHealthCheckFailedError: Any other error on health check
        :raises APIDataNotAvailableError: Dataset is not available at the endpoint
        """
        self.base_url = url
        self.url = f"{url}/v1/{dataset}"
        self.interpolation = interpolation

        if not skip_checks:
            logger.debug("Doing server health check")
            try:
                resp = requests.get(f"{self.base_url}/health")
            except requests.exceptions.ConnectionError as e:
                raise APIHealthCheckFailedError(str(e))
            if resp.status_code != 200:
                raise APIHealthCheckFailedError(resp.text)

            logger.debug("Doing dataset check")
            resp = requests.get(f"{self.base_url}/datasets")
            if resp.status_code != 200:
                raise APIHealthCheckFailedError(resp.text)
            datasets = [ds["name"] for ds in resp.json()["results"]]
            if dataset not in datasets:
                raise APIDataNotAvailableError("Dataset %s not available" % dataset)

    def get_elevation_data(
        self,
        input_coordinates: list[tuple[float, float]],
        split_requests: None | int = None,
    ) -> list[float]:
        """Send a post request to the api endoint to query elevation data
        for the passed input coordinates

        :param input_coordinates: list of latitude, longitude tuples for which the
            elevation should be determined.

        :param split_requests: Optionally split request into multiple requires to get
            around size restrictions, defaults to None
        :raises APIResponseError: Any none 200 response form the endpoint

        :returns: A list of Elevations for the passed coordinates.
        """
        logger.debug("Getting elevation data")
        if split_requests is None:
            split_input_coord = [input_coordinates]
        else:
            split_input_coord = [
                input_coordinates[i : i + split_requests]
                for i in range(0, len(input_coordinates), split_requests)
            ]

        ret_elevations = []
        for coords in split_input_coord:
            locations = ""
            for latitude, longitude in coords:
                locations += f"{latitude},{longitude}|"

            locations = locations[:-1]
            resp = requests.post(
                self.url,
                data={
                    "locations": locations,
                    "interpolation": self.interpolation,
                },
            )

            if resp.status_code == 200:
                result_data = resp.json()
                for res in result_data["results"]:
                    ret_elevations.append(res["elevation"])

            else:
                raise APIResponseError(resp.text)

        return ret_elevations


@final
class OpenElevationEnhancer(ElevationEnhancer):
    """Use the/a OpenElevation API (https://open-elevation.com) to enhance a GPX track
    with elevation information."""

    def __init__(self, url: str = "https://api.open-elevation.com") -> None:
        """
        Setup the enhancer via the url of the rest api of open-elevation

        :param url: URL of the API gateway
        """
        self.url = f"{url}/api/v1/lookup"

        self.headers: Mapping[str, str] = CaseInsensitiveDict()
        self.headers["Accept"] = "application/json"
        self.headers["Content-Type"] = "application/json"

    def get_elevation_data(
        self, input_coordinates: list[tuple[float, float]]
    ) -> list[float]:
        """
        Send a POST request to the Open-Elevation API specified in the init.

        :param input_coordinates: list of latitude, longitude tuples for which the
            elevation should be determined.

        :returns: A list of Elevations for the passed coordinates.
        """
        data: Dict = {"locations": []}
        for latitude, longitude in input_coordinates:
            data["locations"].append({"latitude": latitude, "longitude": longitude})

        resp = requests.post(self.url, headers=self.headers, data=json.dumps(data))

        if resp.status_code == 200:
            result_data = resp.json()
            ret_elevations = []
            for res in result_data["results"]:
                ret_elevations.append(float(res["elevation"]))

            return ret_elevations
        else:
            raise APIResponseError(resp.text)


def get_enhancer(name: EnhancerType) -> Type[Enhancer]:
    """Get a Enhance object for a specific enpoint by passing a distinct name

    :param name: Name of enhancer. Chose OpenTopoElevation or OpenElevation
    :raises NotImplementedError: If an invalid name is passed
    :return: An Enhancer object
    """
    if name == EnhancerType.OPENTOPOELEVATION:
        return OpenTopoElevationEnhancer
    elif name == EnhancerType.OPENELEVATION:
        return OpenElevationEnhancer
    else:
        raise NotImplementedError("Can not return Enhancer for name %s" % name)
